@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python -i
 
 # Fabric - Pythonic remote deployment tool.
 # Copyright (C) 2008  Christian Vest Hansen
@@ -22,6 +22,7 @@ import os
 import os.path
 import re
 import signal
+import subprocess
 import sys
 import threading
 import time
@@ -31,10 +32,11 @@ import datetime
 try:
     import paramiko as ssh
 except ImportError:
-    print("ERROR: paramiko is a required module. Please install it.")
+    print("Error: paramiko is a required module. Please install it:")
+    print("  $ sudo easy_install paramiko")
     exit(1)
 
-__version__ = '0.0.1'
+__version__ = '0.0.3'
 __author__ = 'Christian Vest Hansen'
 __author_email__ = 'karmazilla@gmail.com'
 __url__ = 'https://savannah.nongnu.org/projects/fab/'
@@ -52,22 +54,49 @@ ENV = {
     'fab_mode':'fanout',
     'fab_port':22,
     'fab_user':None,
-    'fab_password':None,
     'fab_pkey':None,
     'fab_key_filename':None,
     'fab_new_host_key':'accept',
     'fab_shell':'/bin/bash -l -c "%s"',
-    # TODO: make fab_timestamp UTC
-    'fab_timestamp':datetime.datetime.now().strftime('%F_%H-%M-%S'),
+    'fab_timestamp':datetime.datetime.utcnow().strftime('%F_%H-%M-%S'),
     'fab_debug':False,
 }
-COMMANDS = {} # populated by load() and _load_std_commands()
+
 CONNECTIONS = []
-OPERATIONS = {} # populated by _load_operations_helper_map()
+COMMANDS = {}
+OPERATIONS = {}
+STRATEGIES = {}
+_LAZY_FORMAT_SUBSTITUTER = re.compile(r'\$\((?P<var>\w+?)\)')
+
+#
+# Helper decorators:
+#
+def new_registering_decorator(registry):
+    def registering_decorator(first_arg=None):
+        if callable(first_arg):
+            registry[first_arg.__name__] = first_arg
+            return first_arg
+        else:
+            def sub_decorator(f):
+                registry[first_arg] = f
+                return f
+            return sub_decorator
+    return registering_decorator
+command = new_registering_decorator(COMMANDS)
+operation = new_registering_decorator(OPERATIONS)
+strategy = new_registering_decorator(STRATEGIES)
+
+def run_per_host(op_fn):
+    def wrapper(*args, **kvargs):
+        if not CONNECTIONS:
+            _connect()
+        _on_hosts_do(op_fn, *args, **kvargs)
+    return wrapper
 
 #
 # Standard fabfile operations:
 #
+@operation
 def set(**variables):
     """Set a number of Fabric environment variables.
     
@@ -90,23 +119,26 @@ def set(**variables):
         else:
             ENV[k] = v
 
-def get(name):
+@operation
+def get(name, otherwise=None):
     """Get the value of a given Fabric environment variable.
     
-    If the variable isn't found, then this operation returns None.
+    If the variable isn't found, then this operation returns the
+    value of the 'otherwise' parameter, which is None unless set.
     
     """
-    return name in ENV and ENV[name] or None
+    return name in ENV and ENV[name] or otherwise
 
+@operation
 def require(var, **kvargs):
     """Make sure that a certain environmet variable is available.
     
     The 'var' parameter is a string that names the variable to check for.
     Two other optional kvargs are supported:
-     - 'used_for' is a string that gets injected into, and then printed, as
-       something like this string: "This variable is used for %s".
-     - 'provided_by' is a list of strings that name commands which the user
-       can run in order to satisfy the requirement.
+        * 'used_for' is a string that gets injected into, and then printed, as
+          something like this string: "This variable is used for %s".
+        * 'provided_by' is a list of strings that name commands which the user
+          can run in order to satisfy the requirement.
     
     If the required variable is not found in the current environment, then the
     operation is stopped and Fabric halts.
@@ -125,13 +157,17 @@ def require(var, **kvargs):
         + "' variable.") % ENV
     )
     if 'used_for' in kvargs:
-        print("This variable is used for %s" % kvargs['used_for'])
+        print("This variable is used for %s" % _lazy_format(kvargs['used_for']))
     if 'provided_by' in kvargs:
         print("Get the variable by running one of these commands:")
-        print('\t' + ('\n\t'.join(kvargs['provided_by'])))
+        to_s = lambda obj: getattr(obj, '__name__', str(obj))
+        provided_by = [to_s(obj) for obj in kvargs['provided_by']]
+        print('\t' + ('\n\t'.join(provided_by)))
     exit(1)
 
-def put(localpath, remotepath, **kvargs):
+@operation
+@run_per_host
+def put(host, client, env, localpath, remotepath, **kvargs):
     """Upload a file to the current hosts.
     
     The 'localpath' parameter is the relative or absolute path to the file on
@@ -140,27 +176,86 @@ def put(localpath, remotepath, **kvargs):
     fab_hosts, and relative paths are relative to the fab_user's home
     directory.
     
+    May take an additional 'fail' keyword argument with one of these values:
+        * ignore - do nothing on failure
+        * warn - print warning on failure
+        * abort - terminate fabric on failure
+    
     Example:
         put('bin/project.zip', '/tmp/project.zip')
     
     """
-    if not CONNECTIONS: _connect()
-    _on_hosts_do(_put, localpath, remotepath)
+    localpath = _lazy_format(localpath, env)
+    remotepath = _lazy_format(remotepath, env)
+    if not os.path.exists(localpath):
+        return False
+    ftp = client.open_sftp()
+    print("[%s] put: %s -> %s" % (host, localpath, remotepath))
+    ftp.put(localpath, remotepath)
+    return True
 
-def run(cmd, **kvargs):
+@operation
+@run_per_host
+def download(host, client, env, remotepath, localpath, **kvargs):
+    """Download a file from the remote hosts.
+    
+    The 'remotepath' parameter is the relative or absolute path to the files
+    to download from the fab_hosts. The 'localpath' parameter will be suffixed
+    with the individual hostname from which they were downloaded, and the
+    downloaded files will then be stored in those respective paths.
+    
+    May take an additional 'fail' keyword argument with one of these values:
+        * ignore - do nothing on failure
+        * warn - print warning on failure
+        * abort - terminate fabric on failure
+    
+    Example:
+        set(fab_hosts=['node1.cluster.com','node2.cluster.com'])
+        download('/var/log/server.log', 'server.log')
+    
+    The above code will produce two files on your local system, called
+    "server.log.node1.cluster.com" and "server.log.node2.cluster.com"
+    respectively.
+    
+    """
+    ftp = client.open_sftp()
+    localpath = _lazy_format(localpath) + '.' + host
+    remotepath = _lazy_format(remotepath)
+    print("[%s] download: %s <- %s" % (host, localpath, remotepath))
+    ftp.get(remotepath, localpath)
+    return True
+
+@operation
+@run_per_host
+def run(host, client, env, cmd, **kvargs):
     """Run a shell command on the current fab_hosts.
     
     The provided command is executed with the permisions of fab_user, and the
     exact execution environ is determined by the fab_shell variable.
     
+    May take an additional 'fail' keyword argument with one of these values:
+        * ignore - do nothing on failure
+        * warn - print warning on failure
+        * abort - terminate fabric on failure
+    
     Example:
         run("ls")
     
     """
-    if not CONNECTIONS: _connect()
-    _on_hosts_do(_run, cmd)
+    cmd = env['fab_shell'] % _lazy_format(cmd, env).replace('"', '\\"')
+    if not _confirm_proceed('run', host, kvargs):
+        return False # TODO: should we return False in fail??
+    print("[%s] run: %s" % (host, cmd))
+    stdin, stdout, stderr = client.exec_command(cmd)
+    out_th = _start_outputter("[%s] out" % host, stdout)
+    err_th = _start_outputter("[%s] err" % host, stderr)
+    out_th.join()
+    err_th.join()
+    return True
 
-def sudo(cmd, **kvargs):
+@operation
+@run_per_host
+def sudo(host, client, env, cmd, **kvargs):
     """Run a sudo (root privileged) command on the current hosts.
     
     The provided command is executed with root permisions, provided that
@@ -168,25 +263,59 @@ def sudo(cmd, **kvargs):
     environ is determined by the fab_shell variable - the 'sudo' part is
     injected into this variable.
     
+    May take an additional 'fail' keyword argument with one of these values:
+        * ignore - do nothing on failure
+        * warn - print warning on failure
+        * abort - terminate fabric on failure
+    
     Example:
         sudo("install_script.py")
     
     """
-    if not CONNECTIONS: _connect()
-    _on_hosts_do(_sudo, cmd)
+    cmd = _lazy_format(cmd, env)
+    real_cmd = env['fab_shell'] % ("sudo -S " + cmd.replace('"', '\\"'))
+    cmd = env['fab_debug'] and real_cmd or cmd
+    if not _confirm_proceed('sudo', host, kvargs):
+        return False # TODO: should we return False in fail??
+    print("[%s] sudo: %s" % (host, cmd))
+    stdin, stdout, stderr = client.exec_command(real_cmd)
+    stdin.write(env['fab_password'])
+    stdin.write('\n')
+    stdin.flush()
+    out_th = _start_outputter("[%s] out" % host, stdout)
+    err_th = _start_outputter("[%s] err" % host, stderr)
+    out_th.join()
+    err_th.join()
+    return True
 
+@operation
 def local(cmd, **kvargs):
     """Run a command locally.
     
     This operation is essentially 'os.system()' except that variables are
     expanded prior to running.
     
+    May take an additional 'fail' keyword argument with one of these values:
+        * ignore - do nothing on failure
+        * warn - print warning on failure
+        * abort - terminate fabric on failure
+    
     Example:
         local("make clean dist")
     
     """
-    os.system(_lazy_format(cmd))
+    final_cmd = _lazy_format(cmd)
+    print("[localhost] run: " + final_cmd)
+    retcode = subprocess.call(final_cmd, shell=True)
+    if retcode != 0:
+        failcode = _get_failcode(kvargs)
+        if failcode > 1:
+            print("Warning: failed to execute command:")
+            print("\t" + final_cmd)
+        if failcode > 2:
+            exit(1)
 
+@operation
 def local_per_host(cmd, **kvargs):
     """Run a command locally, for every defined host.
     
@@ -194,6 +323,8 @@ def local_per_host(cmd, **kvargs):
     with this operation, the command is executed (and have its variables
     expanded) for each host in fab_hosts.
     
+    This operation is defined to take the same keyword arguments as local().
+    
     Example:
         local_per_host("scp -i login.key stuff.zip $(fab_host):stuff.zip")
     
@@ -201,27 +332,40 @@ def local_per_host(cmd, **kvargs):
     _check_fab_hosts()
     for host in ENV['fab_hosts']:
         ENV['fab_host'] = host
-        cur_cmd = _lazy_format(cmd)
-        os.system(cur_cmd)
+        local(cmd, **kvargs)
 
-def load(filename):
+@operation
+def load(filename, **kvargs):
     """Load up the given fabfile.
     
     This loads the fabfile specified by the 'filename' parameter into fabric
     and make its commands and other functions available in the scope of the 
     current fabfile.
     
+    May take an additional 'fail' keyword argument with one of these values:
+        * ignore - do nothing on failure
+        * warn - print warning on failure
+        * abort - terminate fabric on failure
+    
     Example:
         load("conf/production-settings.py")
     
     """
-    execfile(filename)
-    for name, obj in locals().items():
-        if not name.startswith('_') and isinstance(obj, types.FunctionType):
-            COMMANDS[name] = obj
-        if not name.startswith('_'):
-            __builtins__[name] = obj
+    failcode = _get_failcode(kvargs)
+    if os.path.exists(filename):
+        execfile(filename)
+        for name, obj in locals().items():
+            if not name.startswith('_') and isinstance(obj, types.FunctionType):
+                COMMANDS[name] = obj
+            if not name.startswith('_'):
+                __builtins__[name] = obj
+    elif failcode > 1:
+        print("Warning: Cannot load file '%s'." % filename)
+        print("No such file in your current directory.")
+        if failcode > 2:
+            exit(1)
 
+@operation
 def upload_project(**kvargs):
     """Uploads the current project directory to the connected hosts.
     
@@ -239,17 +383,18 @@ def upload_project(**kvargs):
     """
     tar_file = "/tmp/fab.%(fab_timestamp)s.tar" % ENV
     cwd_name = os.getcwd().split(os.sep)[-1]
-    local("tar -czf %s ." % tar_file)
-    put(tar_file, cwd_name + ".tar.gz")
-    local("rm -f " + tar_file)
-    run("tar -xzf " + cwd_name)
-    run("rm -f " + cwd_name + ".tar.gz")
+    local("tar -czf %s ." % tar_file, **kvargs)
+    put(tar_file, cwd_name + ".tar.gz", **kvargs)
+    local("rm -f " + tar_file, **kvargs)
+    run("tar -xzf " + cwd_name, **kvargs)
+    run("rm -f " + cwd_name + ".tar.gz", **kvargs)
 
 #
 # Standard Fabric commands:
 #
+@command("help")
 def _help(**kvargs):
-    """Display usage help message to the console, or help for a given command.
+    """Display Fabric usage help, or help for a given command.
     
     You can provide help with a parameter and get more detailed help for a
     specific command. For instance, to learn more about the list command, you
@@ -261,10 +406,11 @@ def _help(**kvargs):
     more about. For instance, to learn more about the 'run' operation, you
     could run 'fab help:op=run'.
     
+    Lastly, you can also learn more about a certain strategy with the 'strg'
+    and 'strategy' parameters: 'fab help:strg=rolling'.
+    
     """
     if kvargs:
-        if not OPERATIONS:
-            _load_operations_helper_map()
         for k, v in kvargs.items():
             if k in COMMANDS:
                 _print_help_for_in(k, COMMANDS)
@@ -272,24 +418,28 @@ def _help(**kvargs):
                 _print_help_for_in(k, OPERATIONS)
             elif k in ['op', 'operation']:
                 _print_help_for_in(kvargs[k], OPERATIONS)
+            elif k in ['strg', 'strategy']:
+                _print_help_for_in(kvargs[k], STRATEGIES)
             else:
                 _print_help_for(k, None)
     else:
         print("""Fabric is a simple pythonic remote deployment tool.
-        
-        Type 'fab list' to get a list of available commands.
-        Type 'fab help:help' to get more information on how to use the built in
-        help.
-        
-        """)
+    
+    Type 'fab list' to get a list of available commands.
+    Type 'fab help:help' to get more information on how to use the built in
+    help.
+    
+    """)
 
+@command("list")
 def _list_commands(**kvargs):
     """Display a list of commands with descriptions.
     
     By default, the list command prints a list of available commands, with a
     short description (if one is available). However, the list command can also
     print a list of available operaions if you provide it with the 'ops' or
-    'operations' parameter: 'fab list:ops'.
+    'operations' parameters, or it can print strategies with the 'strgs' and
+    'strategies' parameters.
     
     """
     if kvargs:
@@ -298,23 +448,28 @@ def _list_commands(**kvargs):
                 print("Available commands are:")
                 _list_objs(COMMANDS)
             elif k in ['ops', 'operations']:
-                print("Available operations:")
+                print("Available operations are:")
                 _list_objs(OPERATIONS)
+            elif k in ['strgs', 'strategies']:
+                print("Available strategies are:")
+                _list_objs(STRATEGIES)
             else:
                 print("Don't know how to list '%s'." % k)
                 print("Try one of these instead:")
                 print(_indent('\n'.join([
                     'cmds', 'commands',
                     'ops', 'operations',
+                    'strgs', 'strategies',
                 ])))
                 exit(1)
     else:
         print("Available commands are:")
         _list_objs(COMMANDS)
 
+@command("license")
 def _license():
     "Display the Fabric distribution license text."
-    print """		    GNU GENERAL PUBLIC LICENSE
+    print("""		    GNU GENERAL PUBLIC LICENSE
    TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
 
   0. This License applies to any program or other work which contains
@@ -511,11 +666,12 @@ to ask for permission.  For software which is copyrighted by the Free
 Software Foundation, write to the Free Software Foundation; we sometimes
 make exceptions for this.  Our decision will be guided by the two goals
 of preserving the free status of all derivatives of our free software and
-of promoting the sharing and reuse of software generally."""
+of promoting the sharing and reuse of software generally.""")
 
+@command("warranty")
 def _warranty():
     "Display warranty information for the Fabric software."
-    print """			    NO WARRANTY
+    print("""			    NO WARRANTY
 
   BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
 FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
@@ -535,8 +691,9 @@ OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
 TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
 YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
 PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
-POSSIBILITY OF SUCH DAMAGES."""
+POSSIBILITY OF SUCH DAMAGES.""")
 
+@command("set")
 def _set(**kvargs):
     """Set a Fabric variable.
     
@@ -546,6 +703,7 @@ def _set(**kvargs):
     for k, v in kvargs.items():
         ENV[k] = (v % ENV)
 
+@command("shell")
 def _shell(**kvargs):
     "Start an interactive shell connection to the specified hosts."
     def lines():
@@ -563,41 +721,44 @@ def _shell(**kvargs):
         else:
             run(line)
 
+#
+# Standard strategies:
+#
+@strategy("fanout")
+def _fanout_strategy(fn, *args, **kvargs):
+    """A strategy that executes on all hosts in parallel.
+    
+    THIS STRATEGY IS CURRENTLY BROKEN!
+    
+    """
+    threads = []
+    success = True
+    for host, client in CONNECTIONS:
+        env = dict(ENV)
+        env['fab_host'] = host
+        def functor():
+            success = success and fn(host, client, env, *args, **kvargs)
+        thread = threading.Thread(None, functor)
+        thread.setDaemon(True)
+        thread.start()
+        threads.append(thread)
+    for thread in threads:
+        thread.join()
+    return success
+
+@strategy("rolling")
+def _rolling_strategy(fn, *args, **kvargs):
+    """One-at-a-time fail-fast strategy."""
+    for host, client in CONNECTIONS:
+        env = dict(ENV)
+        env['fab_host'] = host
+        if not fn(host, client, env, *args, **kvargs):
+            return False
+    return True
+
 #
 # Internal plumbing:
 #
-def _pick_fabfile():
-    "Figure out what the fabfile is called."
-    choise = 'fabfile'
-    alternatives = ['Fabfile', 'fabfile.py', 'Fabfile.py']
-    for alternative in alternatives:
-        if os.path.exists(alternative):
-            choise = alternative
-            break
-    return choise
-
-def _load_std_commands():
-    "Loads up the standard commands such as help, list, warranty and license."
-    COMMANDS['help'] = _help
-    COMMANDS['list'] = _list_commands
-    COMMANDS['warranty'] = _warranty
-    COMMANDS['license'] = _license
-    COMMANDS['set'] = _set
-    COMMANDS['shell'] = _shell
-
-def _load_operations_helper_map():
-    "Loads up the standard operations in OPERATIONS so 'help' can query them."
-    OPERATIONS['set'] = set
-    OPERATIONS['get'] = get
-    OPERATIONS['put'] = put
-    OPERATIONS['run'] = run
-    OPERATIONS['sudo'] = sudo
-    OPERATIONS['require'] = require
-    OPERATIONS['local'] = local
-    OPERATIONS['local_per_host'] = local_per_host
-    OPERATIONS['load'] = load
-    OPERATIONS['upload_project'] = upload_project
-
 def _indent(text, level=4):
     "Indent all lines in text with 'level' number of spaces, default 4."
     return '\n'.join(((' ' * level) + line for line in text.splitlines()))
@@ -636,13 +797,13 @@ def _check_fab_hosts():
 def _connect():
     "Populate CONNECTIONS with (hostname, client) tuples as per fab_hosts."
     _check_fab_hosts()
-    signal.signal(signal.SIGINT, _trap_sigint)
+    signal.signal(signal.SIGINT, lambda: _disconnect() and exit(0))
     if 'fab_password' not in ENV:
         print(_lazy_format("Logging into the following hosts as $(fab_user):"))
         print(_indent('\n'.join(ENV['fab_hosts'])))
         ENV['fab_password'] = getpass.getpass()
     else:
-        print("WARNING: Putting your password in a fabfile is a bad idea.")
+        print("Warning: Putting your password in a fabfile is a bad idea.")
     port = int(ENV['fab_port'])
     username = ENV['fab_user']
     password = ENV['fab_password']
@@ -669,81 +830,57 @@ def _disconnect():
         client.close()
     CONNECTIONS = []
 
-def _trap_sigint(signal, frame):
-    "Trap ctrl-c and make sure we disconnect everything."
-    _disconnect()
-    exit(0)
-
 def _lazy_format(string, env=ENV):
     "Do recursive string substitution of ENV vars - both lazy and earger."
-    suber = re.compile(r'\$\((?P<var>\w+?)\)')
     def replacer_fn(match):
         var = match.group('var')
         if var in env:
             return _lazy_format(env[var] % env, env)
         else:
             return match.group(0)
-    return re.sub(suber, replacer_fn, string % env)
+    return re.sub(_LAZY_FORMAT_SUBSTITUTER, replacer_fn, string % env)
 
-def _on_hosts_do(fn, *args):
+def _on_hosts_do(fn, *args, **kvargs):
     """Invoke the given function with hostname and client parameters in
     accord with the current fac_mode strategy.
     
-    fn should be of type:
-        (str:hostname, paramiko.SSHClient:clinet) -> bool:success
+    fn should be a callable taking these parameters:
+        hostname : str
+        client : paramiko.SSHClient
+        *args
+        **kvargs
     
     """
     strategy = ENV['fab_mode']
-    if strategy == 'fanout':
-        threads = []
-        for host, client in CONNECTIONS:
-            env = dict(ENV)
-            env['fab_host'] = host
-            thread = threading.Thread(None, lambda: fn(host, client, env, *args))
-            thread.setDaemon(True)
-            thread.start()
-            threads.append(thread)
-        for thread in threads:
-            thread.join()
-    elif strategy == 'rolling':
-        for host, client in CONNECTIONS:
-            env = dict(ENV)
-            env['fab_host'] = host
-            fn(host, client, env, *args)
+    if strategy in STRATEGIES:
+        strategy_fn = STRATEGIES[strategy]
+        successful = strategy_fn(fn, *args, **kvargs)
+        if not successful:
+            print("Operation failed: " + fn.__name__)
+            exit(1)
     else:
         print("Unsupported fab_mode: %s" % strategy)
         print("Supported modes are: fanout, rolling")
         exit(1)
 
-def _put(host, client, env, localpath, remotepath):
-    ftp = client.open_sftp()
-    localpath = _lazy_format(localpath, env)
-    remotepath = _lazy_format(remotepath, env)
-    print("[%s] put: %s -> %s" % (host, localpath, remotepath))
-    ftp.put(localpath, remotepath)
+def _confirm_proceed(exec_type, host, kvargs):
+    if 'confirm' in kvargs:
+        infotuple = (exec_type, host, _lazy_format(kvargs['confirm']))
+        question = "Confirm %s for host %s: %s [yN] " % infotuple
+        answer = raw_input(question)
+        return answer in 'yY'
+    return True
 
-def _run(host, client, env, cmd):
-    cmd = _lazy_format(cmd, env)
-    real_cmd = env['fab_shell'] % cmd.replace('"', '\\"')
-    print("[%s] run: %s" % (host, (env['fab_debug'] and real_cmd or cmd)))
-    stdin, stdout, stderr = client.exec_command(real_cmd)
-    out_th = _start_outputter("[%s] out" % host, stdout)
-    err_th = _start_outputter("[%s] err" % host, stderr)
-    out_th.join()
-    err_th.join()
-
-def _sudo(host, client, env, cmd):
-    cmd = _lazy_format(cmd, env)
-    real_cmd = env['fab_shell'] % ("sudo -S " + cmd.replace('"', '\\"'))
-    print("[%s] sudo: %s" % (host, (env['fab_debug'] and real_cmd or cmd)))
-    stdin, stdout, stderr = client.exec_command(real_cmd)
-    stdin.write(env['fab_password'])
-    stdin.write('\n')
-    stdin.flush()
-    out_th = _start_outputter("[%s] out" % host, stdout)
-    err_th = _start_outputter("[%s] err" % host, stderr)
-    out_th.join()
-    err_th.join()
+def _get_failcode(kvarg_map, default='abort'):
+    codes = {
+        'ignore': 1,
+        'warn': 2,
+        'abort': 3,
+    }
+    if 'fail' in kvarg_map:
+        return codes[kvarg_map['fail']]
+    else:
+        return codes[default]
 
 def _start_outputter(prefix, channel):
     def outputter():
@@ -756,6 +893,31 @@ def _start_outputter(prefix, channel):
     thread.start()
     return thread
 
+def _pick_fabfile():
+    "Figure out what the fabfile is called."
+    guesses = ['fabfile','Fabfile', 'fabfile.py', 'Fabfile.py']
+    options = filter(os.path.exists, guesses)
+    if options:
+        return options[0]
+    else:
+        return guesses[0] # load() will barf for us...
+
+def _parse_args(args):
+    cmds = []
+    for cmd in args:
+        cmd_name = cmd
+        cmd_args = None
+        if cmd.find(':') != -1:
+            cmd_name, cmd_args = cmd.split(':', 1)
+        if cmd_args is not None:
+            cmd_arg_kvs = {}
+            for cmd_arg_kv in cmd_args.split(','):
+                k, _, v = cmd_arg_kv.partition('=')
+                cmd_arg_kvs[k] = (v % ENV)
+            cmd_args = cmd_arg_kvs
+        cmds.append((cmd_name, cmd_args))
+    return cmds
+
 def _validate_commands(args):
     for cmd in args:
         if cmd.find(':') != -1:
@@ -788,7 +950,6 @@ def _execute_commands(args):
 def main(args):
     try:
         print(__greeter__ % ENV)
-        _load_std_commands()
         fabfile = _pick_fabfile()
         load(fabfile)
         _validate_commands(args)
